#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-, :pep:`612`-, and :pep:`646`-compliant **type parameter
utilities** (i.e., low-level callables generically handling :pep:`484`-compliant
type variables, pep:`612`-compliant parameter specifications, and
:pep:`646`-compliant type variable tuples).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484612646Exception
from beartype.typing import TypeVar
from beartype._cave._cavefast import (
    HintPep646692UnpackedType,
    HintPep484612646TypeArgPackedTypes,
)
from beartype._data.hint.sign.datahintsigns import (
    HintSignPep646TypeVarTupleUnpacked)
from beartype._data.typing.datatyping import (
    Pep484612646TypeArgPacked,
    Pep484612646TypeArgUnpacked,
    TypeException,
)
from beartype._data.typing.datatypingport import (
    Hint,
    TypeIs,
)
from beartype._util.hint.pep.utilpepget import get_hint_pep_args

# ....................{ RAISERS                            }....................
#FIXME: Unit test us up, please.
def die_unless_hint_pep484612646_typearg_unpacked(
    # Mandatory parameters.
    hint: Pep484612646TypeArgUnpacked,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep484612646Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception unless the passed type hint is an **unpacked type
    parameter** (i.e., :pep:`484`-compliant type variable,
    :pep:`612`-compliant unpacked parameter specification, or
    :pep:`646`-compliant unpacked type variable tuple).

    Parameters
    ----------
    hint : Pep484612646TypeArgUnpacked
        Type hint to be validated.
    exception_cls : Type[Exception], default: BeartypeDecorHintPep484612646Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep484612646Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If this type hint is *not* an unpacked type parameter.
    '''

    # If this hint is *NOT* an unpacked type parameter, raise an exception.
    if not is_hint_pep484612646_typearg_unpacked(hint):  # pyright: ignore
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not exception subclass.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} not '
            f'unpacked type parameter (i.e., '
            f'PEP 484 "typing.TypeVar(...)" type variable or '
            f'PEP 646 "*typing.TypeVarTuple(...)" unpacked type variable tuple'
            f').'
        )
    # Else, this hint is an unpacked type parameter.


#FIXME: Unit test us up, please.
def die_unless_hint_pep484612646_typearg_packed(
    # Mandatory parameters.
    hint: Pep484612646TypeArgPacked,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep484612646Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception unless the passed type hint is a **packed type
    parameter** (i.e., :pep:`484`-compliant type variable, pep:`612`-compliant
    parameter specification, or :pep:`646`-compliant type variable tuples).

    Parameters
    ----------
    hint : Pep484612646TypeArgPacked
        Type hint to be validated.
    exception_cls : Type[Exception], default: BeartypeDecorHintPep484612646Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep484612646Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If this type hint is *not* a packed type parameter.
    '''

    # If this hint is *NOT* a packed type parameter, raise an exception.
    if not is_hint_pep484612646_typearg_packed(hint):  # pyright: ignore
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not exception subclass.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} not '
            f'type parameter (i.e., '
            f'PEP 484 "typing.TypeVar(...)" type variable, or '
            f'PEP 612 "typing.ParamSpec(...)" parameter specification, or '
            f'PEP 646 "typing.TypeVarTuple(...)" unpacked type variable tuple'
            f').'
        )
    # Else, this hint is an unpacked type parameter.

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_hint_pep484612646_typearg_unpacked(
    hint: Hint) -> TypeIs[Pep484612646TypeArgUnpacked]:  # pyright: ignore
    '''
    :data:`True` only if the passed type hint is a **unpacked type parameter**
    (i.e., :pep:`484`-compliant type variable, :pep:`612`-compliant unpacked
    parameter specification, or :pep:`646`-compliant unpacked type variable
    tuple).

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this hint is a packed type parameter.
    '''

    # If this hint is a PEP 484-compliant type variable, this hint is a type
    # parameter. In this case, immediately return true.
    if isinstance(hint, TypeVar):
        return True
    # Else, this hint is *NOT* a PEP 484-compliant type variable. This hint
    # could still be a PEP 646-compliant unpacked type variable tuple, though.
    #
    # If this hint is *NOT* a PEP 646-compliant or 692-compliant unpacked hint,
    # this hint *CANNOT* by extension be a PEP 646-compliant unpacked type
    # variable tuple. In this case, return false.
    elif not isinstance(hint, HintPep646692UnpackedType):
        return False
    # Else, this hint is a PEP 646-compliant or 692-compliant unpacked hint and
    # thus could be a PEP 646-compliant unpacked type variable tuple.

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

    # Sign uniquely identifying this hint.
    hint_sign = get_hint_pep_sign_or_none(hint)  # pyright: ignore

    # Return true only if this sign uniquely identifies this hint to be a PEP
    # 646-compliant unpacked type variable tuple.
    return hint_sign is HintSignPep646TypeVarTupleUnpacked


#FIXME: Unit test us up, please.
def is_hint_pep484612646_typearg_packed(
    hint: Hint) -> TypeIs[Pep484612646TypeArgPacked]:  # pyright: ignore
    '''
    :data:`True` only if the passed type hint is a **packed type parameter**
    (i.e., :pep:`484`-compliant type variable, pep:`612`-compliant parameter
    specification, or :pep:`646`-compliant type variable tuples).

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this hint is a packed type parameter.
    '''

    # Return true only if this hint is a packed type parameter.
    return isinstance(hint, HintPep484612646TypeArgPackedTypes)

# ....................{ GETTERS                            }....................
def get_hint_pep484612646_typearg_packed_name(
    # Mandatory parameters.
    hint: Pep484612646TypeArgPacked,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep484612646Exception,
    exception_prefix: str = '',
) -> str:
    '''
    Unqualified basename of the passed **packed type parameter** (i.e.,
    :pep:`484`-compliant type variable, pep:`612`-compliant parameter
    specification, or :pep:`646`-compliant type variable tuples).

    Parameters
    ----------
    hint : Pep484612646TypeArgPacked
        Type parameter to be inspected.
    exception_cls : Type[Exception], default: BeartypeDecorHintPep484612646Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintForwardRefException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    str
        Unqualified basename of this type parameter.

    Raises
    ------
    exception_cls
        If this object is *not* a type parameter.
    '''

    # If this hint is *NOT* a packed type parameter, raise an exception.
    die_unless_hint_pep484612646_typearg_packed(
        hint=hint,  # pyright: ignore
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this hint is a packed type parameter.

    # Return the name of this type parameter. Thankfully, *ALL* type parameters
    # generically conform to this simplistic API. We give minor praise.
    return hint.__name__  # type: ignore[union-attr]

# ....................{ PACKERS                            }....................
def pack_hint_pep484612646_typearg_unpacked(
    # Mandatory parameters.
    hint: Pep484612646TypeArgUnpacked,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep484612646Exception,
    exception_prefix: str = '',
) -> Pep484612646TypeArgPacked:
    '''
    **Packed type parameter** (i.e., :pep:`484`-compliant type variable,
    pep:`612`-compliant parameter specification, or :pep:`646`-compliant type
    variable tuples) underlying the passed **unpacked type parameter** (i.e.,
    :pep:`484`-compliant type variable, pep:`612`-compliant unpacked parameter
    specification, or :pep:`646`-compliant unpacked type variable tuples).

    Specifically, if the passed unpacked type parameter is:

    * A :pep:`484`-compliant type variable, this function returns the same type
      variable unmodified.
    * A pep:`612`-compliant unpacked parameter specification, this function
      returns the lower-level packed parameter specification underlying this
      unpacked parameter specification (e.g., from ``*P`` to merely ``P``).
    * A pep:`646`-compliant unpacked type variable tuple, this function
      returns the lower-level packed type variable tuple underlying this
      unpacked type variable tuple (e.g., from ``*Ts`` to merely ``Ts``).

    Parameters
    ----------
    hint : Pep484612646TypeArgPacked
        Type parameter to be inspected.
    exception_cls : Type[Exception], default: BeartypeDecorHintPep484612646Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintForwardRefException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    Pep484612646TypeArgPacked
        Packed type parameter underlying this unpacked type parameter.

    Raises
    ------
    exception_cls
        If this object is *not* a type parameter.
    '''

    # If this hint is *NOT* an unpacked type parameter, raise an exception.
    die_unless_hint_pep484612646_typearg_unpacked(
        hint=hint,  # pyright: ignore
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this hint is an unpacked type parameter.

    # If this hint is a PEP 484-compliant type variable, this type parameter's
    # packed and unpacked forms are equivalent. In this case, return this type
    # variable as is.
    if isinstance(hint, TypeVar):
        return hint
    # Else, this hint is *NOT* a PEP 484-compliant type variable. By
    # elimination, this hint *MUST* be either:
    # * A PEP 612-compliant unpacked parameter specification.
    # * A PEP 646-compliant unpacked type variable tuple.

    # Tuple of the zero or more child hints subscripting this unpacked type
    # parameter.
    hint_args = get_hint_pep_args(hint)

    # If this unpacked type parameter is *NOT* subscripted by exactly one child
    # hint, raise an exception.
    if len(hint_args) != 1:
        raise exception_cls(
            f'{exception_prefix}unpacked type parameter {repr(hint)} invalid '
            f'(i.e., unpacks {len(hint_args)} type parameters rather than '
            f'1 type parameter).'
        )
    # Else, this unpacked type parameter is subscripted by one child hint.

    # This child hint.
    hint_packed = hint_args[0]

    # If this hint is *NOT* a packed type parameter, raise an exception.
    die_unless_hint_pep484612646_typearg_packed(
        hint=hint_packed,  # pyright: ignore
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this hint is an packed type parameter.

    # Return this packed type parameter.
    return hint_packed



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype callable code object utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.utilfunc.utilfunccodeobj` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from functools import wraps

# ....................{ DATA                               }....................
#FIXME: Refactor as follows for sanity, please
#* Shift into the "beartype_test.a00_unit.data.data_type" submodule.
#* Rename this class to something suitably generic -- say, "ClassWithWrappers".
#  Or perhaps just shift these methods into an existing class of that submodule?
#* Import this class into tests below.

# Arbitrary pure-Python class.
class OfStagnantWaters(object):
    # Arbitrary pure-Python wrappee method.
    def altar(self, sword: str, and_pen: str) -> str:
        return sword + and_pen

    # Arbitrary pure-Python wrappee method.
    @wraps(altar)
    def and_give_us_manners_virtue_freedom_power(
        self, *args, **kwargs) -> str:
        return self.altar(*args, **kwargs)

# ....................{ TESTS                              }....................
def test_get_func_codeobj() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` function.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._cave._cavefast import CallableCodeObjectType
    from beartype._util.func.utilfunccodeobj import get_func_codeobj
    from beartype_test.a00_unit.data.data_type import (
        function,
        sync_generator_factory,
        wrapper_isomorphic,
    )
    from pytest import raises

    # ..................{ UNWRAPPING ~ false                 }..................
    # Instance of that class.
    fireside = OfStagnantWaters()

    # Assert this tester accepts pure-Python callables.
    function_codeobj = get_func_codeobj(function)
    assert isinstance(function_codeobj, CallableCodeObjectType)
    assert isinstance(get_func_codeobj(fireside.altar), CallableCodeObjectType)

    # Assert this tester accepts pure-Python generators.
    assert isinstance(
        get_func_codeobj(sync_generator_factory), CallableCodeObjectType)

    # Assert this tester also accepts code objects.
    assert get_func_codeobj(function_codeobj) is function_codeobj

    # Assert this tester rejects C-based callables.
    with raises(_BeartypeUtilCallableException):
        get_func_codeobj(iter)

    # ..................{ UNWRAPPING ~ true                  }..................
    # Assert this tester accepts pure-Python wrappee callables.
    function_codeobj = get_func_codeobj(
        func=function, is_unwrap=True)
    have_forfeited_their_ancient_english_dower = get_func_codeobj(
        func=fireside.altar, is_unwrap=True)
    assert isinstance(
        function_codeobj, CallableCodeObjectType)
    assert isinstance(
        have_forfeited_their_ancient_english_dower, CallableCodeObjectType)

    # Assert this tester accepts pure-Python wrapper callables.
    assert (
        get_func_codeobj(func=wrapper_isomorphic, is_unwrap=True) is
        get_func_codeobj(func=function, is_unwrap=False)
    )
    assert get_func_codeobj(
        func=fireside.and_give_us_manners_virtue_freedom_power,
        is_unwrap=True
    ) is get_func_codeobj(func=fireside.altar, is_unwrap=False)

    # Assert this tester also accepts code objects.
    assert get_func_codeobj(
        function_codeobj, is_unwrap=True) is function_codeobj

    # Assert this tester rejects C-based callables.
    with raises(_BeartypeUtilCallableException):
        get_func_codeobj(iter, is_unwrap=True)


def test_get_func_codeobj_or_none() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj_or_none`
    function.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._cave._cavefast import CallableCodeObjectType
    from beartype._util.func.utilfunccodeobj import get_func_codeobj_or_none
    from beartype_test.a00_unit.data.data_type import (
        function,
        sync_generator_factory,
        wrapper_isomorphic,
    )

    # ..................{ UNWRAPPING ~ false                 }..................
    # Instance of that class.
    fireside = OfStagnantWaters()

    # Assert this tester accepts pure-Python non-generator callables.
    function_codeobj = get_func_codeobj_or_none(function)
    assert isinstance(function_codeobj, CallableCodeObjectType)
    assert isinstance(
        get_func_codeobj_or_none(fireside.altar), CallableCodeObjectType)

    # Assert this tester accepts pure-Python generators.
    assert isinstance(
        get_func_codeobj_or_none(sync_generator_factory),
        CallableCodeObjectType,
    )

    # Assert this tester also accepts code objects.
    assert get_func_codeobj_or_none(function_codeobj) is function_codeobj

    # Assert this tester rejects C-based builtins.
    assert get_func_codeobj_or_none(iter) is None

    # ..................{ UNWRAPPING ~ true                  }..................
    # Assert this tester accepts pure-Python wrappee callables.
    function_codeobj = get_func_codeobj_or_none(
        func=function, is_unwrap=True)
    have_forfeited_their_ancient_english_dower = get_func_codeobj_or_none(
        func=fireside.altar, is_unwrap=True)
    assert isinstance(function_codeobj, CallableCodeObjectType)
    assert isinstance(
        have_forfeited_their_ancient_english_dower, CallableCodeObjectType)

    # Assert this tester accepts pure-Python wrapper callables.
    assert (
        get_func_codeobj_or_none(wrapper_isomorphic, is_unwrap=True) is
        get_func_codeobj_or_none(function, is_unwrap=False)
    )
    assert get_func_codeobj_or_none(
        func=fireside.and_give_us_manners_virtue_freedom_power,
        is_unwrap=True,
    ) is get_func_codeobj_or_none(fireside.altar, is_unwrap=False)

    # Assert this tester also accepts code objects.
    assert get_func_codeobj_or_none(
        func=function_codeobj, is_unwrap=True) is function_codeobj

    # Assert this tester rejects C-based builtins.
    assert get_func_codeobj_or_none(iter, is_unwrap=True) is None



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **NumPy globals** (i.e., global constants pertaining to the
third-party :mod:`numpy` package).

This private submodule is *not* intended for importation by downstream callers.

Caveats
-------
**Never unconditionally import this submodule from global scope.** Only
conditionally import this submodule after validating that :mod:`numpy` is
importable under the active Python interpreter: e.g.,

.. code-block:: python

   from beartype._util.module.utilmodtest import is_module
   if is_module('numpy'):
       from beartype._data.api import dataapinumpy
       ...
'''

#FIXME: Currently unused, but preserved for posterity. What can you do? *shrug*
# # ....................{ IMPORTS                            }....................
# from numpy import (
#     bool_,
#     bytes_,
#     str_,
# )
#
# # ....................{ DICTS                              }....................
# NUMPY_DTYPE_SIMPLE_TO_BUILTIN_TYPE = {
#     bool_: bool,
#     bytes_: bytes,
#     str_: str,
# }
# '''
# Dictionary mapping from **simple NumPy dtypes** (i.e., dtypes *not* constrained
# to a predefined bitsize) to corresponding builtin types.
# '''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Least Recently Used (LRU) cache** utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: The current "CacheLruStrong" implementation is overly low-level and
#thus fundamentally *THREAD-UNSAFE.* The core issue here is that the current
#approach encourages callers to perform thread-unsafe logic resembling:
#   if key not in lru_dict:  # <-- if a context switch happens here, bad stuff
#       lru_dict[key] = value
#
#For thread-safety, the entire "CacheLruStrong" class *MUST* be rethought along
#the manner of the comparable "utilmapbig.CacheUnboundedStrong" class. Notably:
#* "CacheLruStrong" class should *NOT* directly subclass "dict" but instead
#  simply contain a "_dict" instance.
#* Thread-unsafe dunder methods (particularly the "__setitem__" method) should
#  probably *NOT* be defined at all. Yeah, we know.
#* A new CacheLruStrong.cache_entry() method resembling the existing
#  CacheUnboundedStrong.cache_entry() method should be declared.
#* Indeed, we should (arguably) declare a new "CacheStrongABC" base class to
#  provide a common API here -- trivializing switching between different
#  caching strategies implemented by concrete subclasses.

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCacheLruException
from beartype.typing import Hashable
from threading import Lock

# ....................{ CLASSES                            }....................
class CacheLruStrong(dict):
    '''
    **Thread-safe strong Least Recently Used (LRU) cache** (i.e., mapping
    limited to some maximum capacity of strongly referenced arbitrary keys
    mapped onto strongly referenced arbitrary values, whose methods are
    guaranteed to behave thread-safely).

    Design
    ------
    Cache implementations typically employ weak references for safety.
    Employing strong references invites memory leaks by preventing objects
    *only* referenced by the cache (cache-only objects) from being
    garbage-collected. Nonetheless, this cache intentionally employs strong
    references to persist these cache-only objects across calls to callables
    decorated with :func:`beartype.beartype`. In theory, caching an object
    under a weak reference would result in immediate garbage-collection as,
    with no external strong referents, the object would get collected with all
    other short-lived objects in the first generation (i.e., generation 0).

    Note that:

    * The equivalent LRU cache employing weak references to keys and/or values
      may be trivially implemented by swapping this classes inheritance from
      the builtin :class:`dict` to either of the builtin
      :class:`weakref.WeakKeyDictionary` or
      :class:`weakref.WeakValueDictionary`.
    * The standard example of a cache-only object is a container iterator
      (e.g., :meth:`dict.items`).

    Attributes
    ----------
    _size : int
        **Cache capacity** (i.e., maximum number of key-value pairs persisted
        by this cache).
    _lock : Lock
        **Non-reentrant instance-specific thread lock** (i.e., low-level thread
        locking mechanism implemented as a highly efficient C extension,
        defined as an instance variable for non-reentrant reuse by the public
        API of this class). Although CPython, the canonical Python interpreter,
        *does* prohibit conventional multithreading via its Global Interpreter
        Lock (GIL), CPython still coercively preempts long-running threads at
        arbitrary execution points. Ergo, multithreading concerns are *not*
        safely ignorable -- even under CPython.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        '_size',
        '_lock',
    )

    # ..................{ DUNDERS                            }..................
    def __init__(self, size: int) -> None:
        '''
        Initialize this cache to an empty cache with a capacity of this size.

        Parameters
        ----------
        size : int
            **Cache capacity** (i.e., maximum number of key-value pairs held in
            this cache).

        Raises
        ------
        _BeartypeUtilCacheLruException:
            If the capacity is *not* an integer or its a **non-positive
            integer** (i.e. less than 1).
        '''

        super().__init__()

        if not isinstance(size, int):
            raise _BeartypeUtilCacheLruException(
                f'LRU cache capacity {repr(size)} not integer.')
        elif size < 1:
            raise _BeartypeUtilCacheLruException(
                f'LRU cache capacity {size} not positive.')

        self._size = size
        self._lock = Lock()


    def __getitem__(
        self,
        key: Hashable,

        # Superclass methods efficiently localized as default parameters.
        __contains = dict.__contains__,  # pyright: ignore
        __getitem = dict.__getitem__,  # pyright: ignore
        __delitem = dict.__delitem__,  # pyright: ignore
        __pushitem = dict.__setitem__,  # pyright: ignore
    ) -> object:
        '''
        Return an item previously cached under the passed key *or* raise an
        exception otherwise.

        This implementation is *practically* identical to
        :meth:`self.__contains__` except we return an arbitrary object rather
        than a boolean.

        Parameters
        ----------
        key : Hashable
            Arbitrary hashable key to retrieve the cached value of.

        Returns
        -------
        object
            Arbitrary value cached under this key.

        Raises
        ------
        TypeError
            If this key is not hashable.
        KeyError
            If this key isn't cached.
        '''

        with self._lock:
            # Reset this key if it exists.
            if __contains(self, key):
                val = __getitem(self, key)
                __delitem(self, key)
                __pushitem(self, key, val)
                return val

            raise KeyError(f'Key Error: {key}')


    def __setitem__(
        self,
        key: Hashable,
        value: object,

        # Superclass methods efficiently localized as default parameters.
        __contains = dict.__contains__,  # pyright: ignore
        __delitem = dict.__delitem__,  # pyright: ignore
        __pushitem = dict.__setitem__,  # pyright: ignore
        __iter = dict.__iter__,  # pyright: ignore
        __len = dict.__len__,  # pyright: ignore
    ) -> None:
        '''
        Cache this key-value pair while preserving size constraints.

        Parameters
        ----------
        key : Hashable
            Arbitrary hashable key to cache this value to.
        value : object
            Arbitrary value to be cached under this key.

        Raises
        ------
        TypeError
            If this key is not hashable.
        '''

        with self._lock:
            if __contains(self, key):
                __delitem(self, key)
            __pushitem(self, key, value)

            # Prune this cache.
            if __len(self) > self._size:
                __delitem(self, next(__iter(self)))


    def __contains__(
        self,
        key: Hashable,

        # Superclass methods efficiently localized as default parameters.
        __contains = dict.__contains__,  # pyright: ignore
        __getitem = dict.__getitem__,  # pyright: ignore
        __delitem = dict.__delitem__,  # pyright: ignore
        __pushitem = dict.__setitem__,  # pyright: ignore
     ) -> bool:
        '''
        Return a boolean indicating whether this key is cached.

        If this key is cached, this method implicitly refreshes this key by
        popping and pushing this key back onto the top of this cache.

        Parameters
        ----------
        key : Hashable
            Arbitrary hashable key to detect the existence of.

        Returns
        -------
        bool
            :data:`True` only if this key is cached.

        Raises
        ----------
        TypeError
            If this key is unhashable.
        '''

        with self._lock:
            if __contains(self, key):
                val = __getitem(self, key)
                __delitem(self, key)
                __pushitem(self, key, val)
                return True

            return False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-specific **Python module detection** utilities.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype._util.cache.utilcachecall import callable_cached

# ....................{ TESTERS ~ lib                      }....................
@callable_cached
def is_package_sphinx() -> bool:
    '''
    :data:`True` only if a reasonably recent version of Sphinx is importable
    under the active Python interpreter.
    '''

    # Defer test-specific imports.
    from beartype.meta import _LIB_DOCTIME_MANDATORY_VERSION_MINIMUM_SPHINX
    from beartype._util.module.utilmodtest import is_module_version_at_least

    # Return true only if this version of this package is importable.
    return is_module_version_at_least(
        'sphinx', _LIB_DOCTIME_MANDATORY_VERSION_MINIMUM_SPHINX)


@callable_cached
def is_package_typing_extensions() -> bool:
    '''
    :data:`True` only if a reasonably recent version of the third-party
    :mod:`typing_extensions` package is importable under the active Python
    interpreter.
    '''

    # Defer test-specific imports.
    from beartype.meta import (
        _LIB_RUNTIME_OPTIONAL_VERSION_MINIMUM_TYPING_EXTENSIONS)
    from beartype._util.module.utilmodtest import is_module_version_at_least

    # Return true only if this version of this package is importable.
    return is_module_version_at_least(
        'typing_extensions',
        _LIB_RUNTIME_OPTIONAL_VERSION_MINIMUM_TYPING_EXTENSIONS,
    )

# ....................{ TESTERS ~ lib : numpy              }....................
@callable_cached
def is_package_numpy() -> bool:
    '''
    :data:`True` only if a reasonably recent version of NumPy is importable
    under the active Python interpreter.
    '''

    # Defer test-specific imports.
    from beartype.meta import _LIB_RUNTIME_OPTIONAL_VERSION_MINIMUM_NUMPY
    from beartype._util.module.utilmodtest import is_module_version_at_least

    # Return true only if this version of this package is importable.
    return is_module_version_at_least(
        'numpy', _LIB_RUNTIME_OPTIONAL_VERSION_MINIMUM_NUMPY)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **dictionary tester** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.kind.maplike.utilmaptest` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ validators                 }....................
def test_die_if_mappings_two_items_collide() -> None:
    '''
    Test the
    :func:`beartype._util.kind.maplike.utilmaptest.die_if_mappings_two_items_collide`
    validator.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilMappingException
    from beartype._util.kind.maplike.utilmaptest import (
        die_if_mappings_two_items_collide)
    from beartype_test.a00_unit.data.kind.data_kindmap import (
        FAREWELL_O_HIAWATHA,
        THE_SONG_OF_HIAWATHA,
        THE_SONG_OF_HIAWATHA_SINGING_IN_THE_SUNSHINE,
    )
    from pytest import raises

    # Assert this validator raises the expected exception when passed two
    # non-empty mappings containing one or more key-value collisions whose
    # values are all hashable.
    with raises(_BeartypeUtilMappingException):
        die_if_mappings_two_items_collide(
            THE_SONG_OF_HIAWATHA, FAREWELL_O_HIAWATHA)
    with raises(_BeartypeUtilMappingException):
        die_if_mappings_two_items_collide(
            FAREWELL_O_HIAWATHA, THE_SONG_OF_HIAWATHA)

    # Assert this validator raises the expected exception when passed two
    # non-empty mappings containing one or more key-value collisions such that
    # some values of the second mapping are unhashable.
    with raises(_BeartypeUtilMappingException):
        die_if_mappings_two_items_collide(
            THE_SONG_OF_HIAWATHA, THE_SONG_OF_HIAWATHA_SINGING_IN_THE_SUNSHINE)

# ....................{ TESTS ~ testers                    }....................
def test_is_mapping_keys_all() -> None:
    '''
    Test the
    :func:`beartype._util.kind.maplike.utilmaptest.is_mapping_keys_all` tester.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilMappingException
    from beartype._util.kind.maplike.utilmaptest import is_mapping_keys_all
    from beartype_test.a00_unit.data.kind.data_kindmap import (
        THE_SONG_OF_HIAWATHA,
        THE_SONG_OF_HIAWATHA_SINGING_IN_THE_SUNSHINE,
    )

    # Assert this tester returns true when passed a mapping containing all keys
    # in the passed set.
    assert is_mapping_keys_all(
        mapping=THE_SONG_OF_HIAWATHA_SINGING_IN_THE_SUNSHINE,
        keys=THE_SONG_OF_HIAWATHA.keys(),
    ) is True

    # Assert this tester returns false when passed a mapping *NOT* containing
    # all keys in the passed set.
    assert is_mapping_keys_all(
        mapping=THE_SONG_OF_HIAWATHA_SINGING_IN_THE_SUNSHINE,
        keys=THE_SONG_OF_HIAWATHA.keys() | {'As the mist',},
    ) is False


def test_is_mapping_keys_any() -> None:
    '''
    Test the
    :func:`beartype._util.kind.maplike.utilmaptest.is_mapping_keys_any` tester.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilMappingException
    from beartype._util.kind.maplike.utilmaptest import is_mapping_keys_any
    from beartype_test.a00_unit.data.kind.data_kindmap import (
        FAREWELL_O_HIAWATHA)

    # Assert this tester returns true when passed a mapping containing any key
    # in the passed set.
    assert is_mapping_keys_any(
        mapping=FAREWELL_O_HIAWATHA,
        keys={'Thus departed', 'By the shore', 'To the portals'},
    ) is True

    # Assert this tester returns false when passed a mapping containing *NO* key
    # in the passed set.
    assert is_mapping_keys_any(
        mapping=FAREWELL_O_HIAWATHA,
        keys={'By the shore', 'To the portals'},
    ) is False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Fixed list pool** (i.e., submodule whose thread-safe API caches previously
instantiated :class:`list` subclasses constrained to fixed lengths defined at
instantiation time of various lengths for space- and time-efficient reuse
by the :func:`beartype.beartype` decorator across decoration calls).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Consider submitting the "FixedList" type as a relevant StackOverflow
#answer here:
#    https://stackoverflow.com/questions/10617045/how-to-create-a-fix-size-list-in-python
#    https://stackoverflow.com/questions/51558015/implementing-efficient-fixed-size-fifo-in-python

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCachedFixedListException
from beartype.typing import NoReturn
from beartype._util.cache.pool.utilcachepool import KeyPool
from beartype._util.text.utiltextrepr import represent_object

# ....................{ CONSTANTS                          }....................
FIXED_LIST_SIZE_LARGE = 1024
'''
Extremely large length to constrain acquired and released fixed lists to.

This constant is intended to be passed to the :func:`.acquire_fixed_list`
function, which then returns a fixed list of this length suitable for use in
contexts requiring an "extremely large" list -- where "extremely" and "large"
are both subjective but *should* cover 99.9999% of use cases in this codebase.
'''


FIXED_LIST_SIZE_MEDIUM = 256
'''
Reasonably large length to constrain acquired and released fixed lists to.

This constant is intended to be passed to the :func:`.acquire_fixed_list`
function, which then returns a fixed list of this length suitable for use in
contexts requiring a "reasonably large" list -- where "reasonably" and "large"
are both subjective but *should* cover 99.9999% of use cases in this codebase.
'''

# ....................{ CLASSES                            }....................
class FixedList(list):
    '''
    **Fixed list** (i.e., :class:`list` constrained to a fixed length defined
    at instantiation time).**

    A fixed list is effectively a mutable tuple. Whereas a tuple is immutable
    and thus prohibits changes to its contained items, a fixed list is mutable
    and thus *permits* changes to its contained items.

    Design
    ------
    This list enforces this constraint by overriding *all* :class:`list` dunder
    and standard methods that would otherwise modify the length of this list
    (e.g., :meth:`list.__delitem__`, :meth:`list.append`) to instead
    unconditionally raise an :class:`._BeartypeUtilCachedFixedListException`
    exception.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently
    # called @beartype decorations. Slotting has been shown to reduce read and
    # write costs by approximately ~10%, which is non-trivial.
    __slots__ = ()

    # ..................{ INITIALIZER                        }..................
    def __init__(
        self,

        # Mandatory parameters.
        size: int,

        # Optional parameters.
        obj_init: object = None,
    ) -> None:
        '''
        Initialize this fixed list to the passed length and all items of this
        fixed list to the passed initialization object.

        Parameters
        ----------
        size : IntType
            Length to constrain this fixed list to.
        obj_init : object, optional
            **Initialization object** (i.e., object to be unilaterally assigned
            to *all* indices of this fixed list). Defaults to :data:`None`.

        Raises
        ------
        _BeartypeUtilCachedFixedListException
            If this length is either not an integer *or* is but is
            **non-positive** (i.e., is less than or equal to 0).
        '''

        # If this length is *NOT* an integer, raise an exception.
        if not isinstance(size, int):
            raise _BeartypeUtilCachedFixedListException(
                f'Fixed list length {repr(size)} not integer.')
        # Else, this length is an integer.
        #
        # If this length is non-positive, raise an exception.
        elif size <= 0:
            raise _BeartypeUtilCachedFixedListException(
                f'Fixed list length {size} <= 0.')
        # Else, this length is positive.

        # Make it so with the standard Python idiom for preallocating list
        # space -- which, conveniently, is also the optimally efficient means
        # of doing so. See also the timings in this StackOverflow answer:
        #     https://stackoverflow.com/a/10617221/2809027
        super().__init__([obj_init]*size)

    # ..................{ GOOD ~ non-dunders                 }..................
    # Permit non-dunder methods preserving list length but otherwise requiring
    # overriding.

    def copy(self) -> 'FixedList':

        # Nullified fixed list of the same length as this fixed list.
        list_copy = FixedList(len(self))

        # Slice over the nullified contents of this copy with those of this
        # fixed list.
        list_copy[:] = self

        # Return this copy.
        return list_copy

    # ..................{ BAD ~ dunders                      }..................
    # Prohibit dunder methods modifying list length by overriding these methods
    # to raise exceptions.

    def __delitem__(self, index) -> NoReturn:
        raise _BeartypeUtilCachedFixedListException(
            f'{self._label} index {repr(index)} not deletable.')


    def __iadd__(self, value) -> NoReturn:  # type: ignore[misc]
        raise _BeartypeUtilCachedFixedListException(
            f'{self._label} not addable by {represent_object(value)}.')


    def __imul__(self, value) -> NoReturn:  # type: ignore[misc]
        raise _BeartypeUtilCachedFixedListException(
            f'{self._label} not multipliable by {represent_object(value)}.')

    # ..................{ BAD ~ dunders : setitem            }..................
    #FIXME: Great idea, if efficiency didn't particularly matter. Since
    #efficiency is the entire raison d'etre of this class, however, this method
    #has been temporarily and probably permanently disabled. Extensive
    #profiling has shown this single method to substantially cost us elsewhere.
    #Moreover, this method is only relevant in the context of preventing
    #external callers who are *NOT* us from violating class constraints. No
    #external callers exist, though! We are it. Since we know better, we won't
    #violate class constraints by changing fixed list length with slicing.
    #Moreover, it's unlikely we ever even assign list slices anywhere. *sigh*

    # def __setitem__(self, index, value):
    #
    #     # If these parameters indicate an external attempt to change the length
    #     # of this fixed length with slicing, raise an exception.
    #     self._die_if_slice_len_ne_value_len(index, value)
    #
    #     # If this index is a tuple of 0-based indices and slice objects...
    #     if isinstance(index, Iterable):
    #         # For each index or slice in this tuple...
    #         for subindex in index:
    #             # If these parameters indicate an external attempt to change
    #             # the length of this fixed length with slicing, raise an
    #             # exception.
    #             self._die_if_slice_len_ne_value_len(subindex, value)
    #
    #     # Else, this list is either not being sliced or is but is being set to
    #     # an iterable of the same length as that slice. In either case, this
    #     # operation preserves the length of this list and is thus acceptable.
    #     return super().__setitem__(index, value)


    #FIXME: Disabled as currently only called by __setitem__(). *sigh*
    # def _die_if_slice_len_ne_value_len(self, index, value) -> None:
    #     '''
    #     Raise an exception only if the passed parameters when passed to the
    #     parent :meth:`__setitem__` dunder method signify an external attempt to
    #     change the length of this fixed length with slicing.
    #
    #     This function is intended to be called by the :meth:`__setitem__`
    #     dunder method to validate the passed parameters.
    #
    #     Parameters
    #     ----------
    #     index
    #         0-based index, slice object, or tuple of 0-based indices and slice
    #         objects to index this fixed list with.
    #     value
    #         Object to set this index(s) of this fixed list to.
    #
    #     Raises
    #     ----------
    #     _BeartypeUtilCachedFixedListException
    #         If this index is a **slice object** (i.e., :class:`slice` instance
    #         underlying slice syntax) and this value is either:
    #
    #         * **Unsized** (i.e., unsupported by the :func:`len` builtin).
    #         * Sized but has a length differing from that of this fixed list.
    #     '''
    #
    #     # If this index is *NOT* a slice, silently reduce to a noop.
    #     if not isinstance(index, slice):
    #         return
    #     # Else, this index is a slice.
    #     #
    #     # If this value is *NOT* a sized container, raise an exception.
    #     elif not isinstance(value, Sized):
    #         raise _BeartypeUtilCachedFixedListException(
    #             f'{self._label} slice {repr(index)} not settable to unsized '
    #             f'{represent_object(value)}.'
    #         )
    #     # Else, this value is a sized container.
    #
    #     # 0-based first and one-past-the-last indices sliced by this slice.
    #     start, stop_plus_one, _ = index.indices(len(self))
    #
    #     # Number of items of this fixed list sliced by this slice. By
    #     # definition, this is guaranteed to be a non-negative integer.
    #     slice_len = stop_plus_one - start
    #
    #     # Number of items of this sized container to set this slice to.
    #     value_len = len(value)
    #
    #     # If these two lengths differ, raise an exception.
    #     if slice_len != value_len:
    #         raise _BeartypeUtilCachedFixedListException(
    #             f'{self._label} slice {repr(index)} of length {slice_len} not '
    #             f'settable to {represent_object(value)} of differing '
    #             f'length {value_len}.'
    #         )

    # ..................{ BAD ~ non-dunders                  }..................
    # Prohibit non-dunder methods modifying list length by overriding these
    # methods to raise exceptions.

    def append(self, obj) -> NoReturn:
        raise _BeartypeUtilCachedFixedListException(
            f'{self._label} not appendable by {represent_object(obj)}.')


    def clear(self) -> NoReturn:
        raise _BeartypeUtilCachedFixedListException(
            f'{self._label} not clearable.')


    def extend(self, obj) -> NoReturn:
        raise _BeartypeUtilCachedFixedListException(
            f'{self._label} not extendable by {represent_object(obj)}.')


    def pop(self, *args) -> NoReturn:
        raise _BeartypeUtilCachedFixedListException(
            f'{self._label} not poppable.')


    def remove(self, *args) -> NoReturn:
        raise _BeartypeUtilCachedFixedListException(
            f'{self._label} not removable.')

    # ..................{ PRIVATE ~ property                 }..................
    # Read-only properties intentionally prohibiting mutation.

    @property
    def _label(self) -> str:
        '''
        Human-readable representation of this fixed list trimmed to a
        reasonable length.

        This string property is intended to be interpolated into exception
        messages and should probably *not* be called in contexts where
        efficiency is a valid concern.
        '''

        # One-liners for magnanimous pusillanimousness.
        return f'Fixed list {represent_object(self)}'

# ....................{ PRIVATE ~ factories                }....................
_fixed_list_pool = KeyPool(item_maker=FixedList)
'''
Thread-safe **fixed list pool** (i.e., :class:`.KeyPool` singleton caching
previously instantiated :class:`FixedList` instances of various lengths).

Caveats
-------
**Avoid accessing this private singleton externally.** Instead, call the public
:func:`.acquire_fixed_list` and :func:`.release_fixed_list` functions, which
efficiently validate both input *and* output to conform to sane expectations.
'''

# ....................{ (ACQUIRERS|RELEASERS)              }....................
def acquire_fixed_list(size: int) -> FixedList:
    '''
    Acquire an arbitrary **fixed list** (i.e., :class:`list` constrained to a
    fixed length defined at instantiation time) with the passed length.

    Caveats
    -------
    **The contents of this list are arbitrary.** Callers should make *no*
    assumptions as to this list's initial items, but should instead reinitialize
    this list immediately after acquiring this list with standard list slice
    syntax: e.g.,

    .. code-block:: pycon

       >>> from beartype._util.cache.pool.utilcachepoollistfixed import (
       ...     acquire_fixed_list)
       >>> fixed_list = acquire_fixed_list(size=5)
       >>> fixed_list[:] = ('Dirty', 'Deads', 'Done', 'Dirt', 'Cheap',)

    Parameters
    ----------
    size : int
        Length to constrain the fixed list to be acquired to.

    Returns
    -------
    FixedList
        Arbitrary fixed list with this length.

    Raises
    ------
    _BeartypeUtilCachedFixedListException
        If this length is either not an integer *or* is but is **non-positive**
        (i.e., is less than or equal to 0).
    '''
    # Note that the FixedList.__init__() method already validates this "size"
    # parameter to be an integer.

    # Thread-safely acquire a fixed list of this length.
    fixed_list = _fixed_list_pool.acquire(size)
    assert isinstance(fixed_list, FixedList), (
        f'{repr(fixed_list)} not fixed list.')

    # Return this list.
    return fixed_list


def release_fixed_list(fixed_list: FixedList) -> None:
    '''
    Release the passed fixed list acquired by a prior call to the
    :func:`acquire_fixed_list` function.

    Caveats
    -------
    **This list is not safely accessible after calling this function.** Callers
    should make *no* attempts to read, write, or otherwise access this list, but
    should instead nullify *all* variables referring to this list immediately
    after releasing this list (e.g., by setting these variables to the
    :data:`None` singleton *or* by deleting these variables): e.g.,

    .. code-block:: pycon

       >>> from beartype._util.cache.pool.utilcachepoollistfixed import (
       ...     acquire_fixed_list, release_fixed_list)
       >>> fixed_list = acquire_fixed_list(size=7)
       >>> fixed_list[:] = ('If', 'You', 'Want', 'Blood', "You've", 'Got', 'It',)
       >>> release_fixed_list(fixed_list)
       # Either do this...
       >>> fixed_list = None
       # Or do this.
       >>> del fixed_list

    Parameters
    ----------
    fixed_list : FixedList
        Previously acquired fixed list to be released.
    '''
    assert isinstance(fixed_list, FixedList), (
        f'{repr(fixed_list)} not fixed list.')

    # Thread-safely release this fixed list.
    _fixed_list_pool.release(key=len(fixed_list), item=fixed_list)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`525` **type-checking expression snippets** (i.e.,
triple-quoted pure-Python string constants formatted and concatenated together
to dynamically generate boolean expressions type-checking arbitrary objects
against :pep:`525`-compliant asynchronous generator factories).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.code.datacodename import VAR_NAME_PITH_ROOT
from beartype._data.typing.datatyping import CallableStrFormat

# ....................{ CODE                               }....................
#FIXME: Conditionally assign this string to the
#"self.func_wrapper_code_return_prefix" instance variable in the
#BeartypeDecorMeta.reinit() method, please.

# This pure-Python code snippet is *EXTREMELY* inspired by a comparable snippet
# exhibited in the "Formal Semantics" subsection of PEP 380, which standardized
# the "yield from" expression applicable *ONLY* to synchronous generators:
#     https://peps.python.org/pep-0380/#formal-semantics
#
# PEP 380 differs from most PEPs. Code snippets in most PEPs commonly contain
# syntactic and semantic errors; those snippets were *NEVER* actually tested.
# PEP 695 is the canonical example, containing this particular nugget of bad:
#     # A type alias that includes a forward reference
#     type AnimalOrVegetable = Animal | "Vegetable"
#
# Let's charitably assume the "Animal" type exists. Even then, evaluating that
# PEP 695-compliant type alias raises the expected "TypeError". Why? Because
# strings fail to support the "|" operator, of course:
#     >>> type AnimalOrVegetable = Animal | "Vegetable"
#     >>> class Animal(): ...
#     >>> AnimalOrVegetable.__value__
#     TypeError: unsupported operand type(s) for |: 'type' and 'str'
#
# PEPs are thus untrustworthy with respect to code snippets. For safety, it's
# best to charitably assume *ALL* PEP code to be guilty before proven innocent.
#
# PEP 380 differs. The pure-Python code snippet exhibited in the "Formal
# Semantics" subsection of PEP 380 works -- but it doesn't just work. It appears
# to be the best possible pure-Python approximation of the underlying default C
# implementation of the "yield from" expression in CPython. It doesn't appear
# possible to profitably optimize, refactor, or otherwise improve that snippet.
# We can only surmise that Gregory Ewing (the ingenious author of PEP 380)
# implemented "yield from" with that exact pure-Python snippet as a fully
# working proof-of-concept before eventually unrolling his equivalent C
# implementation -- which makes a great deal of sense. Iterative development is
# considerably faster, easier, and more productive in Python than C. It only
# makes sense to perfect the Python implementation before finalizing that in C.
#
# We thus caution against making *ANY* modifications (however well-intended) to
# this code snippet. You may think you are doing the right thing. You probably
# even believe what you think. But you are wrong. This snippet maximally covers
# all possible edge cases, of which there are uncountably many. With only three
# exceptions, this snippet cannot be improved. These exceptions are:
# 1. Unreadability. For unknown reasons, the original snippet in PEP 380
#    intentionally privatized *ALL* local variables used throughout that snippet
#    as single-letter variable names of the form "_{letter}" (e.g., "_m", "_y").
#    Why? No idea. It doesn't promote readability, usability, or debuggability.
#    The only reason to do that is to obfuscate -- an indefensible reason that
#    has no place in PEP standards. We surmise that Gregory Ewing intended to
#    dissuade brave readers from hand-rolling their own pure-Python alternatives
#    in favour of the standard "yield from" expression. That's an indefensible
#    reason, though. Obfuscation only favours those with something to hide. We
#    globally rename *ALL* such local variables with more appropriate
#    "__beartype_"-prefixed camel-cased nomenclature.
# 2. Type-safety. For generality and safety, the original snippet in PEP 380
#    supports operands that are *NOT* guaranteed to be synchronous generators.
#    Notably, it accesses methods necessarily bound to synchronous generators
#    (but not other objects) with one "try: ... except AttributeError: ..."
#    block for each such method access: e.g.,
#        try:
#            _m = _i.close
#        except AttributeError:
#            pass
#        else:
#            _m()
#
#    Although a sensible precaution for the general-case, this isn't that case.
#    The code snippet defined below applies *ONLY* to callables that are
#    guaranteed to create and return asynchronous generators at call time. Since
#    type safety and thus the existence of these methods is safeguarded, We
#    reduce that defensive coding style to trivial method calls.
# 3. Type-checking. Obviously, the original snippet in PEP 380 fails to apply
#    type-checking. Thankfully, doing so is mostly trivial. Thus, we do.
#
# Lastly, note that:
# * Asynchronous generators *CANNOT* return values -- unlike synchronous
#   generators, which may. While the original snippet in PEP 380 handles such
#   returns, the snippet below *CANNOT* and is thus somewhat terser.
# * The trailing comment "# Pretend you didn't see this:" is a cheap yet
#   worthwhile syntactic hack simplifying code generation elsewhere. For
#   simplicity, the parent
#   "beartype._data.code.datacodefunc.CODE_RETURN_CHECK_SUFFIX" code snippet
#   embedding this child snippet via the "{func_return_prefix}" format variable
#   unconditionally appends "{VAR_NAME_PITH_ROOT}" (i.e., "__beartype_pith_0")
#   to this snippet -- which suffices for all kinds of callables *EXCEPT*:
#   * Asynchronous generators, which are prohibited from returning anything.
#   * Synchronous generators whose returns are annotated, which @beartype
#     type-checks by approximating "yield from" with similar pure-Python logic.
CODE_PEP525_RETURN_PREFIX = f'''
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # [BEGIN "async yield from"] What follows is the pure-Python implementation
    # of the "async yield from" expression... if that existed, which it doesn't.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Attempt to...
    try:
        # Prime the inner generator by awaiting the value yielded by iterating
        # the inner generator once. Note that *ALL* generators are intrinsically
        # iterable without needing to call either the iter() or aiter()
        # builtins, both of which simply return the passed generator as is.
        __beartype_agen_yield_pith = await anext({VAR_NAME_PITH_ROOT})
    # If doing so raised a PEP 525-compliant "StopAsyncIteration" exception,
    # the inner generator finished immediately without yielding anything. This
    # is a valid use case. Squelch this exception and silently reduce to a noop.
    except StopAsyncIteration:
        return
    # Else, doing so raised *NO* exception. In this case...
    #
    # Note that this "else:" branch *CANNOT* be merged into the main body of the
    # sibling "try:" block defined above. Doing so would allow the caller to
    # erroneously halt the inner generator by passing a "StopAsyncIteration"
    # exception to the athrow() method of this outer generator. Generators may
    # be prematurely halted *ONLY* by calling the aclose() method.
    else:
        # PEP 525-compliant bidirectional communication loop, shuttling values
        # and exceptions between the caller above and inner generator below.
        while True:
            # Attempt to...
            try:
                # Yield the value previously yielded by the inner generator up
                # to the caller *AND* capture any value sent in from the caller.
                __beartype_agen_send_pith = yield __beartype_agen_yield_pith
            # If the caller threw a PEP 525-compliant "GeneratorExit" exception
            # into this outer generator, the caller instructed this outer
            # generator to prematurely close prior to completion by calling the
            # aclose() method on this outer generator. This is a valid use case.
            except GeneratorExit as exception:
                # Propagate this closure request to the inner generator.
                await {VAR_NAME_PITH_ROOT}.aclose()

                # Re-raise this exception for orthogonality with PEP 380.
                raise
            # If the caller threw an exception into this outer generator by
            # passing this exception to the athrow() method of this outer
            # generator...
            except BaseException as __beartype_agen_exception:
                # Propagate this exception to the inner generator *AND*, if the
                # inner generator also catches this exception and then resumes
                # operation by yielding a value, capture this value.
                #
                # Note that *ONLY* the anext() method is efficiently
                # accessible as a builtin. For unknown reasons, the athrow()
                # method is *NOT* and must instead be looked up explicitly.
                try:
                    __beartype_agen_yield_pith = (
                        await {VAR_NAME_PITH_ROOT}.athrow(
                            __beartype_agen_exception))
                # If doing so raised a PEP 525-compliant "StopAsyncIteration"
                # exception, the inner generator caught this exception and then
                # finished immediately without yielding anything further. This
                # is a valid use case. Squelch this exception and halt looping.
                except StopAsyncIteration:
                    return
            # Else, doing so raised *NO* exception. In this case...
            #
            # Note that this "else:" branch *CANNOT* be merged into the main
            # body of the sibling "try:" block defined above. See above.
            else:
                # Attempt to...
                try:
                    # If the caller did *NOT* send a value into this outer
                    # generator, the caller iterated this outer generator.
                    # Iterate the inner generator *AND* capture the value
                    # yielded in response. This is the common case.
                    if __beartype_agen_send_pith is None:
                        __beartype_agen_yield_pith = await anext(
                            {VAR_NAME_PITH_ROOT})
                    # Else, the caller sent a value into this outer generator.
                    # Propagate this value to the inner generator *AND* capture
                    # the value yielded in response. This is an edge case.
                    #
                    # Note that *ONLY* the anext() method is efficiently
                    # accessible as a builtin. For unknown reasons, the asend()
                    # method is *NOT* and must instead be looked up explicitly.
                    else:
                        __beartype_agen_yield_pith = (
                            await {VAR_NAME_PITH_ROOT}.asend(
                                __beartype_agen_send_pith))
                # If doing so raised a PEP 525-compliant "StopAsyncIteration"
                # exception, the inner generator finished immediately without
                # yielding anything further. This is a valid use case. Squelch
                # this exception and halt looping.
                #
                # Note that this exception handling *CANNOT* be merged into the
                # exception handling performed by the parent "try:" block. The
                # latter handles exceptions thrown into this outer generator by
                # the caller calling either the aclose() or athrow() methods on
                # this outer generator. Catching this "StopAsyncIteration"
                # exception in the parent "try:" block would allow the caller to
                # erroneously halt the inner generator by passing this exception
                # to the athrow() method of this outer generator. Generators may
                # be prematurely halted *ONLY* by calling the aclose() method.
                except StopAsyncIteration:
                    return
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # [END "async yield from"]
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Pretend you didn't see this, bear fam: '''
'''
:pep:`525`-compliant code snippet prefixing the values to be iteratively yielded
to the caller after calling the decorated asynchronous generator factory in the
body of the wrapper function wrapping that factory with type checking.

This pure-Python code snippet safely implements the hypothetical equivalent of
the ``"async yield from "`` expression -- if that expression existed, which it
does not. Although :pep:`525` claims either pure-Python or C-based
implementations of this expression to currently be infeasible, this snippet
trivially proves that to *not* be the case:

    While it is theoretically possible to implement "yield from" support for
    asynchronous generators, it would require a serious redesign of the
    generators implementation.

See Also
--------
https://github.com/beartype/beartype/issues/592#issuecomment-3559076610
    GitHub issue thread strongly inspiring this implementation. In this thread,
    GitHub users @rboredi and @Glinte present a brilliant first-draft
    pure-Python decorator performing the hypothetical syntactic equivalent of
    the ``"async yield from "`` expression. This code snippet owes a
    considerable debt to @rboredi in particular, who would soon go on to
    extrapolate this explosion of elegance into a full-fledged Python package
    named ``future-async-yield-from``. See below.
https://github.com/rbroderi/future-async-yield-from
    Pure-Python package generalizing the above commentary into a general-purpose
    solution applicable throughout the wider Python ecosystem.
'''

# ....................{ FORMATTERS                         }....................
# str.format() methods, globalized to avoid inefficient dot lookups elsewhere.
# This is an absurd micro-optimization. *fight me, github developer community*
CODE_PEP525_RETURN_PREFIX_format: CallableStrFormat = (
    CODE_PEP525_RETURN_PREFIX.format)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **sequence factories** (i.e., low-level functions creating and
returning various kinds of sequences, both mutable and immutable).
'''

# ....................{ IMPORTS                            }....................
from collections.abc import (
    Sequence as SequenceABC,
)

# ....................{ FACTORIES                          }....................
def make_stack(sequence: SequenceABC) -> list:
    '''
    **Stack** (i.e., efficiently poppable list of all items in the passed
    sequence, intentionally reordered in reverse order such that the first item
    of this sequence is the last item of this list).

    Parameters
    ----------
    sequence : SequenceABC
        Sequence to be coerced into a stack.

    Returns
    -------
    List
        Stack coerced from this sequence.
    '''
    assert isinstance(sequence, SequenceABC), f'{repr(sequence)} not sequence.'

    # Stack coerced from this sequence.
    stack: list = None  # type: ignore[assignment]

    # If this sequence is also a list...
    if isinstance(sequence, list):
        # Non-destructively reverse this list into a reversed copy of this list.
        # Note that this is well-known to be the most efficient means of
        # reversing an existing list into a newly reversed list. See also:
        #     https://stackoverflow.com/a/3705705/2809027
        stack = sequence[::-1]  # behold! the Martian smiley face emoji! ::-1
    # Else, this sequence is *NOT* also a list. In this case...
    else:
        # Coerce this non-list sequence into a list and then destructively
        # reverse this list.
        #
        # Note that this is well-known to be the most efficient means of
        # reversing an arbitrary non-list sequence into a list. See also:
        #     https://stackoverflow.com/a/3705705/2809027
        stack = list(sequence)
        stack.reverse()

    # Return this stack.
    return stack



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-specific **pathname utilities** (i.e., low-level callables inspecting all
categories of pathnames, including basenames, dirnames, filenames, and
filetypes).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ IMPORTS                            }....................
from os.path import sep as DIRNAME_DELIMITER

# ....................{ EXCEPTIONS ~ basename              }....................
#FIXME: Unit test us up, please.
def die_unless_basename(pathname: str) -> None:
    '''
    Raise an exception unless the passed pathname is a **pure basename** (i.e.,
    contains one or more directory separators).

    Parameters
    ----------
    pathname : str
        Pathname to be validated.

    Raises
    ------
    beartype_test._util.pytroar.BeartypeTestPathException
        If this pathname is *not* a pure basename.

    See Also
    --------
    :func:`.is_basename`
        Further details.
    '''

    # Defer test-specific imports.
    from beartype_test._util.pytroar import BeartypeTestPathException

    # If this pathname is *NOT* a pure basename, raise an exception.
    if not is_basename(pathname):
        raise BeartypeTestPathException(
            f'Pathname "{pathname}" not pure basename '
            f'(i.e., contains one or more "{DIRNAME_DELIMITER}" '
            f'directory separator characters).'
        )

# ....................{ TESTERS ~ basename                 }....................
#FIXME: Unit test us up, please.
def is_basename(pathname: str) -> bool:
    '''
    :data:`True` only if the passed pathname is a **pure basename** (i.e.,
    contains no directory separators and hence no directory components).

    Parameters
    ----------
    pathname : str
        Pathname to be tested.

    Returns
    -------
    bool
        :data:`True` only if this pathname is a pure basename.
    '''
    assert isinstance(pathname, str), f'{repr(pathname)} not string.'

    # One-liners justify college education. *OR DO THEY*!?!?
    return DIRNAME_DELIMITER not in pathname



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Callable wrapper utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.utilfunc.utilfuncwrap` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ all                        }....................
def test_unwrap_func_all() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfuncwrap.unwrap_func_all` function.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.func.utilfuncwrap import unwrap_func_all
    from functools import wraps

    # ....................{ CALLABLES                      }....................
    def in_a_station_of_the_metro_line_two():
        '''
        Arbitrary wrappee callable.
        '''

        return 'Petals on a wet, black bough.'


    @wraps(in_a_station_of_the_metro_line_two)
    def in_a_station_of_the_metro():
        '''
        Arbitrary wrapper callable.
        '''

        return (
            'THE apparition of these faces in the crowd;\n' +
            in_a_station_of_the_metro_line_two()
        )

    # ....................{ ASSERTS                        }....................
    # Assert this function returns unwrapped callables unmodified.
    assert unwrap_func_all(in_a_station_of_the_metro_line_two) is (
        in_a_station_of_the_metro_line_two)
    assert unwrap_func_all(iter) is iter

    # Assert this function returns wrapper callables unwrapped.
    assert unwrap_func_all(in_a_station_of_the_metro) is (
        in_a_station_of_the_metro_line_two)


def test_unwrap_func_all_isomorphic() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfuncwrap.unwrap_func_all_isomorphic` function.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.func.utilfuncwrap import unwrap_func_all_isomorphic
    from beartype_test.a00_unit.data.data_type import (
        function_wrappee,
        function_wrapper_nonisomarphic,
        function_wrapper_isomarphic,
        function_wrapper_isomarphic_args,
        function_wrapper_isomarphic_kwargs,
        object_callable_wrapper_isomorphic,
    )

    # ....................{ ASSERTS                        }....................
    # Assert this function returns unwrapped callables unmodified.
    assert unwrap_func_all_isomorphic(function_wrappee) is function_wrappee
    assert unwrap_func_all_isomorphic(iter) is iter

    # Assert this function returns non-isomorphic wrapper callables unmodified.
    assert unwrap_func_all_isomorphic(function_wrapper_nonisomarphic) is (
        function_wrapper_nonisomarphic)

    # Assert this function unwraps isomorphic wrapper callables.
    assert unwrap_func_all_isomorphic(function_wrapper_isomarphic) is (
        function_wrapper_nonisomarphic)
    assert unwrap_func_all_isomorphic(function_wrapper_isomarphic_args) is (
        function_wrapper_nonisomarphic)
    assert unwrap_func_all_isomorphic(function_wrapper_isomarphic_kwargs) is (
        function_wrapper_nonisomarphic)

    # Assert this function unwraps wrapper pseudo-callables.
    object_callable_wrapper_isomorphic_unwrapped = unwrap_func_all_isomorphic(
        func=object_callable_wrapper_isomorphic.__call__,
        wrapper=object_callable_wrapper_isomorphic,
    )
    assert object_callable_wrapper_isomorphic_unwrapped is (
        function_wrapper_nonisomarphic)

# ....................{ TESTS ~ descriptor                 }....................
def test_unwrap_func_class_or_static_method_once() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfuncwrap.unwrap_func_class_or_static_method_once`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableWrapperException
    from beartype._util.func.utilfuncwrap import (
        unwrap_func_class_or_static_method_once)
    from beartype_test.a00_unit.data.data_type import CALLABLES
    from pytest import raises

    # ....................{ CALLABLES                      }....................
    def never_to_be_reclaimed() -> str:
        '''
        Arbitrary pure-Python function.
        '''

        return 'The dwelling-place'


    def becomes_its_spoil() -> str:
        '''
        Arbitrary pure-Python function.
        '''

        return 'The dwelling-place'

    # ....................{ CLASSES                        }....................
    class TheLimitsOfTheDeadAndLivingWorld(object):
        '''
        Arbitrary pure-Python class defining a class method wrapping the
        pure-Python function defined above.
        '''

        of_insects_beasts_and_birds = classmethod(never_to_be_reclaimed)


    class TheirFoodAndTheirRetreatForEverGone(object):
        '''
        Arbitrary pure-Python static defining a static method wrapping the
        pure-Python function defined above.
        '''

        so_much_of_life_and_joy_is_lost = staticmethod(becomes_its_spoil)

    # ....................{ PASS                           }....................
    # Assert this getter returns the expected wrappee when passed a class method
    # descriptor.
    #
    # Note that class method descriptors are *ONLY* directly accessible via the
    # low-level "object.__dict__" dictionary. When accessed as class or instance
    # attributes, class methods reduce to instances of the standard
    # "beartype.cave.MethodBoundInstanceOrClassType" type.
    class_method = TheLimitsOfTheDeadAndLivingWorld.__dict__[
        'of_insects_beasts_and_birds']
    class_method_wrappee = unwrap_func_class_or_static_method_once(class_method)
    assert class_method_wrappee is never_to_be_reclaimed

    # Assert this getter returns the expected wrappee when passed a static
    # method descriptor.
    #
    # Note that static method descriptors are *ONLY* directly accessible via the
    # low-level "object.__dict__" dictionary. When accessed as static or
    # instance attributes, static methods reduce to instances of the standard
    # "beartype.cave.MethodBoundInstanceOrClassType" type.
    static_method = TheirFoodAndTheirRetreatForEverGone.__dict__[
        'so_much_of_life_and_joy_is_lost']
    static_method_wrappee = unwrap_func_class_or_static_method_once(
        static_method)
    assert static_method_wrappee is becomes_its_spoil

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an object
    # that is *NOT* a static method descriptor.
    for some_callable in CALLABLES:
        with raises(_BeartypeUtilCallableWrapperException):
            unwrap_func_class_or_static_method_once(some_callable)


def test_unwrap_func_classmethod_once() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfuncwrap.unwrap_func_classmethod_once`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableWrapperException
    from beartype._util.func.utilfuncwrap import unwrap_func_classmethod_once
    from beartype_test.a00_unit.data.data_type import CALLABLES
    from pytest import raises

    # ....................{ LOCALS                         }....................
    def never_to_be_reclaimed() -> str:
        '''
        Arbitrary pure-Python function.
        '''

        return 'The dwelling-place'


    class TheLimitsOfTheDeadAndLivingWorld(object):
        '''
        Arbitrary pure-Python class defining a class method wrapping the
        pure-Python function defined above.
        '''

        of_insects_beasts_and_birds = classmethod(never_to_be_reclaimed)

    # ....................{ PASS                           }....................
    # Assert this getter returns the expected wrappee when passed a class method
    # descriptor.
    #
    # Note that class method descriptors are *ONLY* directly accessible via the
    # low-level "object.__dict__" dictionary. When accessed as class or instance
    # attributes, class methods reduce to instances of the standard
    # "beartype.cave.MethodBoundInstanceOrClassType" type.
    class_method = TheLimitsOfTheDeadAndLivingWorld.__dict__[
        'of_insects_beasts_and_birds']
    class_method_wrappee = unwrap_func_classmethod_once(class_method)
    assert class_method_wrappee is never_to_be_reclaimed

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an object
    # that is *NOT* a class method descriptor.
    for some_callable in CALLABLES:
        with raises(_BeartypeUtilCallableWrapperException):
            unwrap_func_classmethod_once(some_callable)


def test_unwrap_func_staticmethod_once() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfuncwrap.unwrap_func_staticmethod_once`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableWrapperException
    from beartype._util.func.utilfuncwrap import unwrap_func_staticmethod_once
    from beartype_test.a00_unit.data.data_type import CALLABLES
    from pytest import raises

    # ....................{ LOCALS                         }....................
    def becomes_its_spoil() -> str:
        '''
        Arbitrary pure-Python function.
        '''

        return 'The dwelling-place'

    class TheirFoodAndTheirRetreatForEverGone(object):
        '''
        Arbitrary pure-Python static defining a static method wrapping the
        pure-Python function defined above.
        '''

        so_much_of_life_and_joy_is_lost = staticmethod(becomes_its_spoil)

    # ....................{ PASS                           }....................
    # Assert this getter returns the expected wrappee when passed a static method
    # descriptor.
    #
    # Note that static method descriptors are *ONLY* directly accessible via the
    # low-level "object.__dict__" dictionary. When accessed as static or instance
    # attributes, static methods reduce to instances of the standard
    # "beartype.cave.MethodBoundInstanceOrClassType" type.
    static_method = TheirFoodAndTheirRetreatForEverGone.__dict__[
        'so_much_of_life_and_joy_is_lost']
    static_method_wrappee = unwrap_func_staticmethod_once(static_method)
    assert static_method_wrappee is becomes_its_spoil

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an object
    # that is *NOT* a static method descriptor.
    for some_callable in CALLABLES:
        with raises(_BeartypeUtilCallableWrapperException):
            unwrap_func_staticmethod_once(some_callable)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Python word size** utility unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.py.utilpyword` submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                             }....................
def test_word_size() -> None:
    '''
    Test the :func:`beartype._util.py.utilpyword.WORD_SIZE` constant.
    '''

    # Defer test-specific imports.
    from beartype._util.py.utilpyword import WORD_SIZE

    # Assert the active Python interpreter to be either 32- or 64-bit.
    assert WORD_SIZE in {32, 64}



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype** :pep:`585` **utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep585` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_hint_pep585_builtin(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep585.is_hint_pep585_builtin_subbed`
    function.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of type hint metadata describing sample type hints exercising edge
        cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep585 import (
        is_hint_pep585_builtin_subbed)

    # Assert this tester accepts only PEP 585-compliant type hints.
    for hint_pep_meta in hints_pep_meta:
        assert is_hint_pep585_builtin_subbed(hint_pep_meta.hint) is (
            hint_pep_meta.is_pep585_builtin_subbed)


def test_is_hint_pep585_generic(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep585.is_hint_pep585_generic`
    function.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of type hint metadata describing sample type hints exercising edge
        cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep585 import (
        is_hint_pep585_generic)

    # Assert this tester accepts only PEP 585-compliant generics.
    for hint_pep_meta in hints_pep_meta:
        assert is_hint_pep585_generic(hint_pep_meta.hint) is (
            hint_pep_meta.is_pep585_generic)

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep585_generic_typeargs_packed(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep585.get_hint_pep585_generic_typeargs_packed`
    function.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of type hint metadata describing sample type hints exercising edge
        cases in the :mod:`beartype` codebase.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep585Exception
    from beartype._data.hint.sign.datahintsigns import HintSignTypeVar
    from beartype._util.hint.pep.proposal.pep585 import (
        get_hint_pep585_generic_typeargs_packed)
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
    from pytest import raises

    # ....................{ ASSERTS                        }....................
    # For each PEP-compliant test hint...
    for hint_pep_meta in hints_pep_meta:
        # If this hint is a PEP 585-compliant generic...
        if hint_pep_meta.is_pep585_generic:
            # Tuple of all type variables discovered by this getter.
            hint_typevars = get_hint_pep585_generic_typeargs_packed(hint_pep_meta.hint)
            assert isinstance(hint_typevars, tuple)

            # Assert that all items of this tuple are actually type variables.
            for hint_typevar in hint_typevars:
                assert get_hint_pep_sign_or_none(hint_typevar) is (
                    HintSignTypeVar)

            # If this hint is parametrized by one or more type variables...
            if hint_pep_meta.is_typeargs:
                # Assert that this getter returns one or more type variables.
                assert hint_typevars

                # If the exact type variables parametrizing this hint are known
                # at test time...
                if hint_pep_meta.typeargs_packed:
                    # Assert that this getter returns only these type variables.
                    assert hint_pep_meta.typeargs_packed == hint_typevars
                # Else, the exact type variables parametrizing this hint are
                # unknown at test time. In this case, silently ignore the exact
                # contents of this tuple.
            # Else, this hint is unparametrized by type variables. In this case,
            # assert that this getter returns the empty tuple.
            else:
                assert hint_typevars == ()
        # If this hint is *NOT* a PEP 585-compliant generic, assert that this
        # getter raises an exception.
        else:
            with raises(BeartypeDecorHintPep585Exception):
                get_hint_pep585_generic_typeargs_packed(hint_pep_meta.hint)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **object utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.utilobject` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_object_hashable() -> None:
    '''
    Test the :func:`beartype._util.utilobject.is_object_hashable` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.utilobject import is_object_hashable
    from beartype_test.a00_unit.data.hint.data_hint import (
        NOT_HINTS_HASHABLE,
        NOT_HINTS_UNHASHABLE,
    )

    # Assert this tester accepts unhashable objects.
    for object_hashable in NOT_HINTS_HASHABLE:
        assert is_object_hashable(object_hashable) is True

    # Assert this tester rejects unhashable objects.
    for object_unhashable in NOT_HINTS_UNHASHABLE:
        assert is_object_hashable(object_unhashable) is False

# ....................{ TESTS ~ getter : name              }....................
def test_get_object_basename_scoped() -> None:
    '''
    Test the :func:`beartype._util.utilobject.get_object_basename_scoped`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilObjectNameException
    from beartype._util.utilobject import get_object_basename_scoped
    from beartype_test.a00_unit.data.data_type import (
        CALLABLES,
        closure_factory,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this getter returns the fully-qualified names of non-nested
    # callables unmodified.
    for callable_obj in CALLABLES:
        assert get_object_basename_scoped(callable_obj) == (
            callable_obj.__qualname__)

    # Assert this getter returns the fully-qualified names of closures stripped
    # of meaningless "<locals>." substrings.
    assert get_object_basename_scoped(closure_factory()) == (
        'closure_factory.closure')

    # ....................{ FAIL                           }....................
    # Assert this getter raises "AttributeError" exceptions when passed objects
    # declaring neither "__qualname__" nor "__name__" dunder attributes.
    with raises(_BeartypeUtilObjectNameException):
        get_object_basename_scoped(
            'From the ice-gulfs that gird his secret throne,')


def test_get_object_filename_or_none() -> None:
    '''
    Test the :func:`beartype._util.utilobject.get_object_filename_or_none`
    getter.
    '''

    # Defer test-specific imports.
    from beartype._util.utilobject import get_object_filename_or_none
    from beartype_test.a00_unit.data.data_type import (
        Class,
        function,
    )

    # Assert this getter returns the expected filename for a physical class.
    assert 'data_type' in get_object_filename_or_none(Class)

    # Assert this getter returns the expected filename for a physical callable.
    assert 'data_type' in get_object_filename_or_none(function)

    # Assert this getter returns "None" when passed an object that is neither a
    # class *NOR* callable.
    assert get_object_filename_or_none(Class()) is None


def test_get_object_name() -> None:
    '''
    Test the :func:`beartype._util.utilobject.get_object_name` getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilObjectNameException
    from beartype._util.utilobject import get_object_name
    from beartype_test.a00_unit.data.data_type import function_partial
    from pytest import raises

    # ....................{ CALLABLES                      }....................
    def meet_in_the_vale(and_one_majestic_river: str) -> str:
        '''
        Arbitrary nested function.
        '''

        return and_one_majestic_river
    # print(f'meet_in_the_vale.__module__: {meet_in_the_vale.__module__}')

    # ....................{ PASS                           }....................
    # Assert this getter returns the expected name for a nested function.
    meet_in_the_vale_name = get_object_name(meet_in_the_vale)
    assert meet_in_the_vale_name == (
        'beartype_test.a00_unit.a20_util.a00_core.test_utilobject.'
        'test_get_object_name.meet_in_the_vale'
    )

    # ....................{ FAIL                           }....................
    # Assert this getter raises "AttributeError" exceptions when passed objects
    # declaring neither "__qualname__" nor "__name__" dunder attributes.
    with raises(_BeartypeUtilObjectNameException):
        get_object_name(function_partial)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **strongly unbounded cache** (i.e., mapping of unlimited size from
strongly referenced arbitrary keys onto strongly referenced arbitrary values,
whose methods are guaranteed to behave thread-safely) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Callable,
    Dict,
    Union,
)
from beartype._data.kind.datakindiota import SENTINEL
from collections.abc import Hashable
from contextlib import AbstractContextManager
from threading import Lock

# ....................{ CLASSES                            }....................
#FIXME: Submit back to StackOverflow, preferably under this question:
#    https://stackoverflow.com/questions/1312331/using-a-global-dictionary-with-threads-in-python
class CacheUnboundedStrong(object):
    '''
    **Thread-safe strongly unbounded cache** (i.e., mapping of unlimited size
    from strongly referenced arbitrary keys onto strongly referenced arbitrary
    values, whose methods are guaranteed to behave thread-safely).

    Design
    ------
    Cache implementations typically employ weak references for safety. Employing
    strong references invites memory leaks by preventing objects *only*
    referenced by the cache (cache-only objects) from being garbage-collected.
    Nonetheless, this cache intentionally employs strong references to persist
    these cache-only objects across calls to callables decorated with
    :func:`beartype.beartype`. In theory, caching an object under a weak
    reference would result in immediate garbage-collection; with *no* external
    strong referents, that object would be garbage-collected with all other
    short-lived objects in the first generation (i.e., generation 0).

    This cache intentionally does *not* adhere to standard mapping semantics by
    subclassing a standard mapping API (e.g., :class:`dict`,
    :class:`collections.abc.MutableMapping`). Standard mapping semantics are
    sufficiently low-level as to invite race conditions between competing
    threads concurrently contesting the same instance of this class. For
    example, consider the following standard non-atomic logic for caching a new
    key-value into this cache:

    .. code-block:: python

       if key not in cache:    # <-- If a context switch happens immediately
                               # <-- after entering this branch, bad stuff!
           cache[key] = value  # <-- We may overwrite another thread's work.

    Attributes
    ----------
    _key_to_value : dict[Hashable, object]
        Internal **backing store** (i.e., thread-unsafe dictionary of unlimited
        size mapping from strongly referenced arbitrary keys onto strongly
        referenced arbitrary values).
    _key_to_value_get : Callable
        The :meth:`self._key_to_value.get` method, classified for efficiency.
    _key_to_value_set : Callable
        The :meth:`self._key_to_value.__setitem__` dunder method, classified
        for efficiency.
    _lock : AbstractContextManager
        **Instance-specific thread lock** (i.e., low-level thread locking
        mechanism implemented as a highly efficient C extension, defined as an
        instance variable for non-reentrant reuse by the public API of this
        type). Although CPython, the canonical Python interpreter, *does*
        prohibit conventional multithreading via its Global Interpreter Lock
        (GIL), CPython still coercively preempts long-running threads at
        arbitrary execution points. Ergo, multithreading concerns are *not*
        safely ignorable -- even under CPython.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # @beartype decorations. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        '_key_to_value',
        '_key_to_value_get',
        '_key_to_value_set',
        '_lock',
    )

    # ..................{ INITIALIZER                        }..................
    def __init__(
        self,

        # Optional parameters.
        lock_type: Union[type, Callable[[], object]] = Lock,
    ) -> None:
        '''
        Initialize this cache to an empty cache.

        Parameters
        ----------
        lock_type : Union[type, Callable[[], object]]
            Type of thread-safe lock to internally use. Defaults to
            :class:`Lock` (i.e., the type of the standard non-reentrant lock)
            for efficiency.
        '''

        # Initialize all instance variables.
        self._key_to_value: Dict[Hashable, object] = {}
        self._key_to_value_get = self._key_to_value.get
        self._key_to_value_set = self._key_to_value.__setitem__
        self._lock: AbstractContextManager = lock_type()  # type: ignore[assignment]

    # ..................{ GETTERS                            }..................
    def cache_or_get_cached_value(
        self,

        # Mandatory parameters.
        key: Hashable,
        value: object,

        # Hidden parameters, localized for negligible efficiency.
        _SENTINEL=SENTINEL,
    ) -> object:
        '''
        **Statically** (i.e., non-dynamically, rather than "statically" in the
        different semantic sense of "static" methods) associate the passed key
        with the passed value if this cache has yet to cache this key (i.e., if
        this method has yet to be passed this key) and, in any case, return the
        value associated with this key.

        Parameters
        ----------
        key : Hashable
            **Key** (i.e., arbitrary hashable object) to return the associated
            value of.
        value : object
            **Value** (i.e., arbitrary object) to associate with this key if
            this key has yet to be associated with any value.

        Returns
        -------
        object
            **Value** (i.e., arbitrary object) associated with this key.
        '''
        # assert isinstance(key, Hashable), f'{repr(key)} unhashable.'

        # Thread-safely...
        with self._lock:
            # Value previously cached under this key if any *OR* the sentinel
            # placeholder otherwise.
            value_old = self._key_to_value_get(key, _SENTINEL)

            # If this key has already been cached, return this value as is.
            if value_old is not _SENTINEL:
                return value_old
            # Else, this key has yet to be cached.

            # Cache this key with this value.
            self._key_to_value_set(key, value)

            # Return this value.
            return value


    #FIXME: Unit test us up.
    #FIXME: Generalize to accept a new mandatory "arg: object" parameter and
    #then pass rather than forcefully passing the passed key. \o/
    def cache_or_get_cached_func_return_passed_arg(
        self,

        # Mandatory parameters.
        key: Hashable,
        value_factory: Callable[[object], object],
        arg: object,

        # Hidden parameters, localized for negligible efficiency.
        _SENTINEL=SENTINEL,
    ) -> object:
        '''
        Dynamically associate the passed key with the value returned by the
        passed **value factory** (i.e., caller-defined function accepting this
        key and returning the value to be associated with this key) if this
        cache has yet to cache this key (i.e., if this method has yet to be
        passed this key) and, in any case, return the value associated with this
        key.

        Caveats
        ----------
        **This value factory must not recursively call this method.** For
        efficiency, this cache is internally locked through a non-reentrant
        rather than reentrant thread lock. If this value factory accidentally
        recursively calls this method, the active thread will be indefinitely
        locked. Welcome to the risky world of high-cost efficiency gains.

        Parameters
        ----------
        key : Hashable
            **Key** (i.e., arbitrary hashable object) to return the associated
            value of.
        value_factory : Callable[[object], object]
            **Value factory** (i.e., caller-defined function accepting the
            passed ``arg`` object and dynamically returning the value to be
            associated with this key).
        arg : object
            Arbitrary object to be passed as is to this value factory.

        Returns
        -------
        object
            **Value** (i.e., arbitrary object) associated with this key.
        '''
        # assert isinstance(key, Hashable), f'{repr(key)} unhashable.'
        # assert callable(value_factory), f'{repr(value_factory)} uncallable.'

        # Thread-safely...
        with self._lock:
            # Value previously cached under this key if any *OR* the sentinel
            # placeholder otherwise.
            value_old = self._key_to_value_get(key, _SENTINEL)

            # If this key has already been cached, return this value as is.
            if value_old is not _SENTINEL:
                return value_old
            # Else, this key has yet to be cached.

            # Value created by this factory function, localized for negligible
            # efficiency to avoid the unnecessary subsequent dictionary lookup.
            value = value_factory(arg)

            # Cache this key with this value.
            self._key_to_value_set(key, value)

            # Return this value.
            return value

    # ..................{ CLEARERS                           }..................
    #FIXME: Unit test us up, please.
    def clear(self) -> None:
        '''
        Clear (i.e., empty) this cache.
        '''

        # Thread-safely...
        with self._lock:
            # Clear your head and be at peace, one-liner.
            self._key_to_value.clear()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype exception raiser utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.error.utilerrwarn` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_reissue_warnings_placeholder() -> None:
    '''
    Test the
    :func:`beartype._util.error.utilerrwarn.reissue_warnings_placeholder`
    function.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.error.utilerrwarn import (
        issue_warning,
        reissue_warnings_placeholder,
    )
    from pytest import warns
    from warnings import catch_warnings

    # ....................{ CLASSES                        }....................
    class TheSkyWarning(UserWarning):
        '''
        Arbitrary warning issued below with messages containing placeholder
        substrings replaced by :func:`.reissue_warnings_placeholder`.
        '''

        pass


    class ThePoetWarning(UserWarning):
        '''
        Arbitrary warning issued below with messages containing placeholder
        substrings replaced by :func:`.reissue_warnings_placeholder`.
        '''

        pass

    # ....................{ LOCALS                         }....................
    # Source substring to be hard-coded into the messages of all warnings
    # issued by the low-level callable defined below.
    TEST_SOURCE_STR = '{its_got_bro}'

    # Target substring to globally replace this source substring with.
    TEST_TARGET_STR = 'With his still soul. '

    # ....................{ CALLABLES                      }....................
    def portend_his_brain() -> None:
        '''
        Low-level callable issuing two or more possibly unreadable warnings.
        '''

        # Issue an arbitrary warning message containing one source substring.
        issue_warning(
            cls=TheSkyWarning,
            message=f'{TEST_SOURCE_STR}The sky',
        )

        # Issue an arbitrary warning message containing *NO* source substring.
        issue_warning(
            cls=ThePoetWarning,
            message='the Poet kept mute conference',
        )

        # Issue an empty warning message.
        issue_warning(cls=TheSkyWarning, message='')

        # Issue a non-string warning message. While an unlikely edge case, the
        # standard low-level warnings.warn() function permissively permits this.
        # Moreover, this behaviour coincides with similar permissiveness in
        # exception messages (which are also allowed to be non-strings): e.g.,
        #     >>> from warnings import warn
        #     >>> warn(0xCAFE, UserWarning)
        #     UserWarning: 51966
        #       warn(0xCAFE, UserWarning)
        issue_warning(cls=ThePoetWarning, message=0xBEEFDEAD)


    def portend_while_daylight() -> None:
        '''
        High-level callable calling the low-level callable and reissuing
        unreadable warnings issued by the latter with equivalent readable
        warnings.
        '''

        # With a context manager "catching" *ALL* non-fatal warnings emitted
        # during this logic for subsequent "playrback" below...
        with catch_warnings(record=True) as warnings_issued:
            # Call this low-level callable to emit multiple warnings.
            portend_his_brain()

        # If one or more warnings were issued, reissue these warnings with each
        # placeholder substring replaced by an arbitrary substring.
        if warnings_issued:
            reissue_warnings_placeholder(
                warnings=warnings_issued,
                source_str=TEST_SOURCE_STR,
                target_str=TEST_TARGET_STR,
            )
        # Else, *NO* warnings were issued.

    # ....................{ ASSERT                         }....................
    # Assert this high-level non-memoized callable raises the same type of
    # exception raised by this low-level memoized callable and preserve this
    # exception for subsequent assertion.
    #
    # Note that passing nothing to warns() enables undocumented edge-case
    # behaviour in which warns() captures warnings issued by this block
    # *WITHOUT* validating the types of these warnings. See also:
    #     https://docs.pytest.org/en/4.6.x/warnings.html#recwarn
    with warns() as warnings_info:
        portend_while_daylight()

    # Assert that the expected number of warnings were issued.
    assert len(warnings_info) == 4

    # Assert that the expected types of warnings were issued.
    assert issubclass(warnings_info[0].category, TheSkyWarning)
    assert issubclass(warnings_info[1].category, ThePoetWarning)
    assert issubclass(warnings_info[2].category, TheSkyWarning)
    assert issubclass(warnings_info[3].category, ThePoetWarning)

    # Assert that the expected warning messages were issued.
    assert str(warnings_info[0].message) == f'{TEST_TARGET_STR}The sky'
    assert str(warnings_info[1].message) == 'the Poet kept mute conference'
    assert str(warnings_info[2].message) == ''
    assert warnings_info[3].message.args == (0xBEEFDEAD,)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hookable configuration first callable decorator position
submodule** (i.e., data module defining callables decorated by chains of one or
more decorators into which the :mod:`beartype.beartype` decorator will be
injected as the first decorator, mimicking real-world usage of the
:func:`beartype.claw.beartype_package` import hook from external callers).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintReturnViolation
from beartype.typing import no_type_check
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
from pytest import raises

# ....................{ FUNCTIONS                          }....................
# Validate that the import hook presumably registered by the caller implicitly
# injects the @beartype decorator as the first decorator in *ALL* chains of
# callable decorations.

@no_type_check
def night_followed() -> int:
    '''
    Arbitrary function trivially violating its return type hint.
    '''

    return 'Night followed, clad with stars. On every side'

# Assert that the import hook registered by the caller ignored the
# @no_type_check decorator decorating the callable defined above by forcefully
# injecting the @beartype decorator as the first decorator on that callable.
with raises(BeartypeCallHintReturnViolation):
    night_followed()

# ....................{ CLASSES                            }....................
# Validate that the import hook presumably registered by the caller implicitly
# injects the @beartype decorator as the last decorator in *ALL* chains of
# class decorations.

# If the active Python interpreter targets Python >= 3.11, the standard
# @no_type_check decorator correctly applies itself to classes. In this case...
if IS_PYTHON_AT_LEAST_3_11:
    @no_type_check
    class MoreHorribly(object):
        '''
        Arbitrary class decorated by one or more decorators, which the import
        hook registered by the caller will then respect by continuing to inject
        the :func:`beartype.beartype` decorator as the last decorator.
        '''

        @staticmethod
        def the_multitudinous_streams() -> int:
            '''
            Arbitrary static method trivially violating its return type hint.
            '''

            return 'More horribly the multitudinous streams'

    # Assert that the import hook registered by the caller respected the
    # @no_type_check decorator decorating the class defined above by continuing
    # to inject the @beartype decorator as the last decorator on that class.
    assert MoreHorribly.the_multitudinous_streams() == (
        'More horribly the multitudinous streams')
# Else, the active Python interpreter targets Python <= 3.10. In this case, the
# standard @no_type_check decorator fails to correctly apply itself to classes.
# In this case, simply avoid performing the prior test for simplicity. *shrug*



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Numba** integration tests.

This submodule functionally tests the :mod:`beartype` package against the
third-party Numba package.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('numba')
def test_numba_njit() -> None:
    '''
    Functional test validating that the :mod:`beartype` package raises *no*
    unexpected exceptions when decorating callables also decorated by the
    third-party :func:`numba.njit` decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from numba import njit

    # ....................{ CALLABLES                      }....................
    #FIXME: Note that enabling "cache=True" causes Numba to
    #non-deterministically raise unreadable exceptions resembling:
    #    RuntimeError: cannot cache function
    #    'test_numba_njit.<locals>.had_shone_gleam_stony_orbs_so_from_his_steps':
    #    no locator available for file
    #    '/home/leycec/py/beartype/beartype_test/a90_func/z90_external/a00_numba/test_numba.py'
    #
    #It's possible this issue could be resolved by explicitly removing *ALL*
    #previously compiled bytecode associated with @njit-decorated callables,
    #perhaps by:
    #* Isolating this callable definition to a new
    #  "beartype_test.a90_func.data.external.numba.data_numba" submodule.
    #* Importing and calling below:
    #      from beartype_test.a90_func.data.external import numba
    #      from beartype_test._util.data.pytdataclean import clean_data_subpackage
    #      clean_data_subpackage(numba)
    #
    #      from beartype_test.a90_func.data.external.numba import data_numba
    #
    #Kinda awkward, but should work. There's no need to bother at the moment,
    #though. Simply disabling caching altogether trivially suffices for now.
    @beartype
    @njit(cache=False)
    def had_shone_gleam_stony_orbs_so_from_his_steps() -> bool:
        '''
        Arbitrary callable decorated by both the :func:`beartype.beartype` *and*
        :func:`numba.njit` decorators in that order.
        '''

        return True

    # ....................{ ASSERTS                        }....................
    # Assert that calling the above callable returns the expected value.
    assert had_shone_gleam_stony_orbs_so_from_his_steps() is True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Decidedly Object-Oriented Runtime-checking (DOOR) union type hint classes**
(i.e., :class:`beartype.door.TypeHint` subclasses implementing support
for :pep:`484`-compliant :attr:`typing.Optional` and :attr:`typing.Union` type
hints and :pep:`604`-compliant ``|``-delimited union type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.typing import Iterable

# ....................{ SUBCLASSES                         }....................
class UnionTypeHint(TypeHint):
    '''
    **Union type hint wrapper** (i.e., high-level object encapsulating a
    low-level :pep:`484`-compliant :attr:`typing.Optional` or
    :attr:`typing.Union` type hint *or* :pep:`604`-compliant ``|``-delimited
    union type hint).
    '''

    # ..................{ PRIVATE ~ properties               }..................
    @property
    def _branches(self) -> Iterable[TypeHint]:
        return self._args_wrapped_tuple

    # ..................{ PRIVATE ~ testers                  }..................
    def _is_subhint(self, other: TypeHint) -> bool:

        # Return true only if *EVERY* child type hint of this union is a subhint
        # of at least one other child type hint of the passed other union.
        #
        # Note that this test has O(n**2) time complexity. Although non-ideal,
        # this is also unavoidable. Thankfully, since most real-world unions are
        # subscripted by only a small number of child type hints, this is also
        # mostly ignorable in practice.
        return all(
            # For each child type hint subscripting this union...
            (
                # If that other type hint is itself a union, true only if...
                any(
                    # For at least one other child type hint subscripting that
                    # other union, this child type hint is a subhint of that
                    # other child type hint.
                    this_branch.is_subhint(that_branch)
                    for that_branch in other._branches
                ) if isinstance(other, UnionTypeHint) else
                # Else, that other type hint is *NOT* a union. In this case,
                # true only if this child type hint is a subhint of that other
                # type hint.
                #
                # Note that this is a common edge case. Examples include:
                # * "TypeHint(Union[...]) <= TypeHint(Any)". Although "Any" is
                #   *NOT* a union, *ALL* unions are subhints of "Any".
                # * "TypeHint(Union[A, B]) <= TypeHint(Union[A])" where "A" is
                #   the superclass of "B". Since Python reduces "Union[A]" to
                #   just "A", this is exactly equivalent to the comparison
                #   "TypeHint(Union[A, B]) <= TypeHint(A)". Although "A" is
                #   *NOT* a union, this example clearly demonstrates that a
                #   union may be a subhint of a non-union that is *NOT* "Any" --
                #   contrary to intuition. Examples include:
                #   * "TypeHint(Union[int, bool]) <= TypeHint(Union[int])".
                this_branch.is_subhint(other)
            )
            for this_branch in self._branches
        )


    def _is_subhint_branch(self, branch: TypeHint) -> bool:  # pragma: no cover
        raise NotImplementedError(
            'UnionTypeHint._is_subhint_branch() unsupported.')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-compliant type hint getter** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.utilpepget` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS ~ attr                       }....................
def test_get_hint_pep_args(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_args`
    getter.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPepException
    from beartype.typing import Tuple
    from beartype._util.hint.pep.utilpepget import (
        get_hint_pep_args,
        _HINT_ARGS_EMPTY_TUPLE,
    )
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP
    from pytest import raises

    # ....................{ CLASSES                        }....................
    class TheWholeMammothBrood(object):
        '''
        Arbitrary PEP-noncompliant type defining the ``__args__`` dunder
        attribute to *not* be a tuple.
        '''

        __args__ = 'But one of the whole mammoth-brood still kept'

    # ....................{ PASS                           }....................
    # For each PEP-compliant hint, assert this getter returns...
    for hint_pep_meta in hints_pep_meta:
        # Localize this hint to simplify debugging.
        hint = hint_pep_meta.hint

        # Tuple of all arguments subscripting this hint.
        hint_args = get_hint_pep_args(hint)
        assert isinstance(hint_args, tuple)

        # For subscripted hints, one or more arguments.
        if hint_pep_meta.is_args:
            assert hint_args
        # For non-argumentative hints, *NO* arguments.
        else:
            assert hint_args == ()

    # Assert this getter returns *NO* type variables for non-"typing" hints.
    for not_hint_pep in NOT_HINTS_PEP:
        assert get_hint_pep_args(not_hint_pep) == ()

    # ....................{ PASS ~ tuple                   }....................
    # Explicitly validate that this getter handles both PEP 484- and 585-
    # compliant empty tuples by returning "_HINT_ARGS_EMPTY_TUPLE" as expected.

    # Assert this getter when passed a PEP 484-compliant empty tuple hint
    # returns a tuple containing an empty tuple for disambiguity.
    assert get_hint_pep_args(Tuple[()]) == _HINT_ARGS_EMPTY_TUPLE

    # Assert this getter when passed a PEP 585-compliant empty tuple hint
    # returns a tuple containing an empty tuple for disambiguity.
    assert get_hint_pep_args(tuple[()]) == _HINT_ARGS_EMPTY_TUPLE

    # ....................{ FAIL                           }....................
    # Assert this getter when passed a PEP-noncompliant hint defining the
    # "__args__" dunder attribute to *NOT* be a tuple raises the expected
    # exception.
    with raises(BeartypeDecorHintPepException):
        get_hint_pep_args(TheWholeMammothBrood)


def test_get_hint_pep_typeargs_packed(hints_pep_meta) -> None:
    '''
    Test the :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_typeargs_packed`
    getter.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPepException
    from beartype._util.hint.pep.proposal.pep484612646 import (
        is_hint_pep484612646_typearg_packed)
    from beartype._util.hint.pep.utilpepget import get_hint_pep_typeargs_packed
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP
    from pytest import raises

    # ....................{ CLASSES                        }....................
    class HisSovereigntyAndRuleAndMajesty(object):
        '''
        Arbitrary PEP-noncompliant type defining the ``__parameters__`` dunder
        attribute to *not* be a tuple.
        '''

        __parameters__ = "His sov'reignty, and rule, and majesty;—"

    # ....................{ PASS                           }....................
    # For each PEP-compliant test hint...
    for hint_pep_meta in hints_pep_meta:
        # Tuple of all packed type parameters discovered by this getter.
        hint_typeargs_packed = get_hint_pep_typeargs_packed(hint_pep_meta.hint)
        assert isinstance(hint_typeargs_packed, tuple)

        # Assert all items of this tuple are actually packed type parameters.
        for hint_typearg in hint_typeargs_packed:
            assert is_hint_pep484612646_typearg_packed(hint_typearg) is True

        # If this hint is parametrized by one or more type parameters...
        if hint_pep_meta.is_typeargs:
            # Assert this getter returns one or more type parameters.
            assert hint_typeargs_packed

            # If the exact type parameters parametrizing this hint are known
            # at test time...
            if hint_pep_meta.typeargs_packed:
                # Assert this getter returns only these type parameters.
                assert hint_pep_meta.typeargs_packed == hint_typeargs_packed
            # Else, the exact type parameters parametrizing this hint are unknown
            # at test time. In this case, silently ignore the exact contents of
            # this tuple.
        # Else, this hint is unparametrized by type parameters. In this case,
        # assert this getter returns the empty tuple.
        else:
            assert hint_typeargs_packed == ()

    # Assert this getter returns *NO* type parameters for non-"typing" hints.
    for not_hint_pep in NOT_HINTS_PEP:
        assert get_hint_pep_typeargs_packed(not_hint_pep) == ()

    # ....................{ FAIL                           }....................
    # Assert this getter when passed a PEP-noncompliant hint defining the
    # "__parameters__" dunder attribute to *NOT* be a tuple raises the expected
    # exception.
    with raises(BeartypeDecorHintPepException):
        get_hint_pep_typeargs_packed(HisSovereigntyAndRuleAndMajesty)

# ....................{ TESTS ~ origin : type              }....................
def test_get_hint_pep_type_isinstanceable(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_origin_type_isinstanceable`
    getter.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPepException
    from beartype._util.hint.pep.utilpepget import (
        get_hint_pep_origin_type_isinstanceable)
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP
    from pytest import raises

    # Assert this getter...
    for hint_pep_meta in hints_pep_meta:
        # Returns the expected type origin for all PEP-compliant type hints
        # originating from an origin type.
        if hint_pep_meta.isinstanceable_type is not None:
            assert get_hint_pep_origin_type_isinstanceable(hint_pep_meta.hint) is (
                hint_pep_meta.isinstanceable_type)
        # Raises the expected exception for all other hints.
        else:
            with raises(BeartypeDecorHintPepException):
                get_hint_pep_origin_type_isinstanceable(hint_pep_meta.hint)

    # Assert this getter raises the expected exception for non-PEP-compliant
    # type hints.
    for not_hint_pep in NOT_HINTS_PEP:
        with raises(BeartypeDecorHintPepException):
            get_hint_pep_origin_type_isinstanceable(not_hint_pep)


def test_get_hint_pep_type_isinstanceable_or_none(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_origin_type_isinstanceable_or_none`
    getter.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype._util.hint.pep.utilpepget import (
        get_hint_pep_origin_type_isinstanceable_or_none)
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP

    # Assert this getter returns the expected type origin for all PEP-compliant
    # type hints.
    for hint_pep_meta in hints_pep_meta:
        assert get_hint_pep_origin_type_isinstanceable_or_none(
            hint_pep_meta.hint) is hint_pep_meta.isinstanceable_type

    # Assert this getter returns "None" for non-PEP-compliant type hints.
    for not_hint_pep in NOT_HINTS_PEP:
        assert get_hint_pep_origin_type_isinstanceable_or_none(
            not_hint_pep) is None



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **non-class decoration globals** (i.e., global constants required
by the private :mod:`beartype._decor._nontype.decornontype` submodule to decorate
non-class objects with runtime type-checking).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._decor._nontype._decordescriptor import (
    beartype_descriptor_decorator_builtin_class_or_static_method,
    beartype_descriptor_decorator_builtin_property,
)
from beartype._util.api.external.utilclick import beartype_click_command

# ....................{ MAPPINGS                           }....................
# Note that this dispatch table is effectively untypeable, thanks to the general
# insanity of pure-static type-checkers (e.g., mypy, pyright). We sigh. *sigh*
MODULE_TO_TYPE_NAME_TO_BEARTYPE_DECORATOR = {
    # ....................{ BUILTINS                       }....................
    # Standard builtins globally accessible *WITHOUT* requiring importation.
    'builtins': {
        'classmethod': (
            beartype_descriptor_decorator_builtin_class_or_static_method),
        'staticmethod': (
            beartype_descriptor_decorator_builtin_class_or_static_method),
        'property': (
            beartype_descriptor_decorator_builtin_property),
    },
}
'''
**Beartype callable decorator exact class dispatch table** (i.e., dictionary
mapping the fully-qualified name of each package or module to the unqualified
basename of each type of an object decoratable by the :func:`beartype.beartype`
decorator defined by that package or module to a decorator function decorating
that object with dynamically generated type-checking).

This dispatch table maps to decorator functions with signatures resembling:

.. code-block:: python

   def {func_name}(func: BeartypeableT, **kwargs) -> BeartypeableT:

The first parameter is passed positionally. All remaining parameters are passed
by keyword and intended to be transitively (i.e., eventually) passed on to the
lower-level :func:`.beartype_func` decorator. In other words, all keyword
parameters accepted by :func:`.beartype_func` are also unconditionally accepted
by *all* of these higher-level decorators.

See Also
--------
:data:`._MODULE_TO_SUPERTYPE_NAME_TO_BEARTYPE_DECORATOR`
    Generalization of this dictionary applicable to the entire method-resolution
    order (MRO) of arbitrary objects.
'''


MODULE_TO_SUPERTYPE_NAME_TO_BEARTYPE_DECORATOR = {
    # ....................{ THIRD-PARTY                    }....................
    # Non-standard types declared by external third-party packages.

    # Click, a popular pure-Python framework for CLI- and TUI-based apps.
    'click.core': {
        # Type created and returned by the @click.command() decorator.
        'Command': beartype_click_command,
    }
}
'''
**Beartype callable decorator superclass dispatch table** (i.e., dictionary
mapping the fully-qualified name of each package or module to the unqualified
basename of each superclass in the method-resolution order (MRO) of an object
decoratable by the :func:`beartype.beartype` decorator defined by that package
or module to a decorator function decorating that object with dynamically
generated type-checking).

See Also
--------
:data:`._MODULE_TO_TYPE_NAME_TO_BEARTYPE_DECORATOR`
    Specialization of this dictionary applicable *only* to the exact class of
    arbitrary objects (rather than the full method-resolution order (MRO) of
    arbitrary objects).
'''

# ....................{ METHODS                            }....................
MODULE_TO_TYPE_NAME_TO_BEARTYPE_DECORATOR_get = (
    MODULE_TO_TYPE_NAME_TO_BEARTYPE_DECORATOR.get)
'''
:meth:`dict.get` method bound to the
:data:`._MODULE_TO_TYPE_NAME_TO_BEARTYPE_DECORATOR` global dictionary,
globalized as an attribute to reduce lookup costs elsewhere.
'''


MODULE_TO_SUPERTYPE_NAME_TO_BEARTYPE_DECORATOR_get = (
    MODULE_TO_SUPERTYPE_NAME_TO_BEARTYPE_DECORATOR.get)
'''
:meth:`dict.get` method bound to the
:data:`._MODULE_TO_SUPERTYPE_NAME_TO_BEARTYPE_DECORATOR` global dictionary,
globalized as an attribute to reduce lookup costs elsewhere.
'''
